@@ -1,5 +1,8 @@
 # -*- coding: utf-8 -*-
+"""Sequential model class and model-related utilities.
+"""
 from __future__ import absolute_import
+from __future__ import division
 from __future__ import print_function
 
 import warnings
@@ -149,7 +152,7 @@ def save_model(model, filepath, overwrite=True, include_optimizer=True):
                     for i, (w, val) in enumerate(zip(symbolic_weights,
                                                      weight_values)):
                         # Default values of symbolic_weights is /variable
-                        # for theano and cntk
+                        # for Theano and CNTK
                         if K.backend() == 'theano' or K.backend() == 'cntk':
                             if hasattr(w, 'name'):
                                 if w.name.split('/')[-1] == 'variable':
@@ -392,8 +395,8 @@ class Sequential(Model):
         self._initial_weights = None
 
         # Model attributes.
-        self.inbound_nodes = []
-        self.outbound_nodes = []
+        self._inbound_nodes = []
+        self._outbound_nodes = []
         self.built = False
 
         # Set model name.
@@ -463,13 +466,13 @@ class Sequential(Model):
                 # to the input layer we just created.
                 layer(x)
 
-            if len(layer.inbound_nodes[-1].output_tensors) != 1:
+            if len(layer._inbound_nodes[-1].output_tensors) != 1:
                 raise ValueError('All layers in a Sequential model '
                                  'should have a single output tensor. '
                                  'For multi-output layers, '
                                  'use the functional API.')
 
-            self.outputs = [layer.inbound_nodes[-1].output_tensors[0]]
+            self.outputs = [layer._inbound_nodes[-1].output_tensors[0]]
             self.inputs = topology.get_source_inputs(self.outputs[0])
 
             # We create an input node, which we will keep updated
@@ -493,9 +496,9 @@ class Sequential(Model):
                                 'For multi-output layers, '
                                 'use the functional API.')
             self.outputs = [output_tensor]
-            # update self.inbound_nodes
-            self.inbound_nodes[0].output_tensors = self.outputs
-            self.inbound_nodes[0].output_shapes = [self.outputs[0]._keras_shape]
+            # update self._inbound_nodes
+            self._inbound_nodes[0].output_tensors = self.outputs
+            self._inbound_nodes[0].output_shapes = [self.outputs[0]._keras_shape]
 
         self.layers.append(layer)
         self.built = False
@@ -512,14 +515,14 @@ class Sequential(Model):
         self.layers.pop()
         if not self.layers:
             self.outputs = []
-            self.inbound_nodes = []
-            self.outbound_nodes = []
+            self._inbound_nodes = []
+            self._outbound_nodes = []
         else:
-            self.layers[-1].outbound_nodes = []
+            self.layers[-1]._outbound_nodes = []
             self.outputs = [self.layers[-1].output]
-            # update self.inbound_nodes
-            self.inbound_nodes[0].output_tensors = self.outputs
-            self.inbound_nodes[0].output_shapes = [self.outputs[0]._keras_shape]
+            # update self._inbound_nodes
+            self._inbound_nodes[0].output_tensors = self.outputs
+            self._inbound_nodes[0].output_shapes = [self.outputs[0]._keras_shape]
         self.built = False
 
     def get_layer(self, name=None, index=None):
@@ -565,12 +568,12 @@ class Sequential(Model):
         self.output_layers = self.model.output_layers
         self.output_layers_node_indices = self.model.output_layers_node_indices
         self.output_layers_tensor_indices = self.model.output_layers_tensor_indices
-        self.nodes_by_depth = self.model.nodes_by_depth
-        self.container_nodes = self.model.container_nodes
+        self._nodes_by_depth = self.model._nodes_by_depth
         self.output_names = self.model.output_names
         self.input_names = self.model.input_names
         self._feed_input_names = self.model._feed_input_names
         self._feed_inputs = self.model._feed_inputs
+        self._container_nodes = self.model._container_nodes
 
         # Make sure child model callbacks
         # will call the parent Sequential model.
@@ -622,7 +625,7 @@ class Sequential(Model):
 
     @trainable.setter
     def trainable(self, value):
-        if self.model:
+        if self.built:
             self.model.trainable = value
         self._trainable = value
 
@@ -715,24 +718,24 @@ class Sequential(Model):
             self.build()
         self.model.set_weights(weights)
 
-    def load_weights(self, filepath, by_name=False):
+    def load_weights(self, filepath, by_name=False, skip_mismatch=False, reshape=False):
         if h5py is None:
             raise ImportError('`load_weights` requires h5py.')
-        f = h5py.File(filepath, mode='r')
-        if 'layer_names' not in f.attrs and 'model_weights' in f:
-            f = f['model_weights']
+        with h5py.File(filepath, mode='r') as f:
+            if 'layer_names' not in f.attrs and 'model_weights' in f:
+                f = f['model_weights']
 
-        # Legacy support
-        if legacy_models.needs_legacy_support(self):
-            layers = legacy_models.legacy_sequential_layers(self)
-        else:
-            layers = self.layers
-        if by_name:
-            topology.load_weights_from_hdf5_group_by_name(f, layers)
-        else:
-            topology.load_weights_from_hdf5_group(f, layers)
-        if hasattr(f, 'close'):
-            f.close()
+            # Legacy support
+            if legacy_models.needs_legacy_support(self):
+                layers = legacy_models.legacy_sequential_layers(self)
+            else:
+                layers = self.layers
+            if by_name:
+                topology.load_weights_from_hdf5_group_by_name(f, layers,
+                                                              skip_mismatch=skip_mismatch,
+                                                              reshape=reshape)
+            else:
+                topology.load_weights_from_hdf5_group(f, layers, reshape=reshape)
 
     def save_weights(self, filepath, overwrite=True):
         if h5py is None:
@@ -748,10 +751,9 @@ class Sequential(Model):
         else:
             layers = self.layers
 
-        f = h5py.File(filepath, 'w')
-        topology.save_weights_to_hdf5_group(f, layers)
-        f.flush()
-        f.close()
+        with h5py.File(filepath, 'w') as f:
+            topology.save_weights_to_hdf5_group(f, layers)
+            f.flush()
 
     def compile(self, optimizer, loss,
                 metrics=None,
@@ -799,6 +801,7 @@ class Sequential(Model):
 
         # Raises
             ValueError: In case of invalid arguments for
+                `optimizer`, `loss`, `metrics` or `sample_weight_mode`.
 
         # Example
             ```python
@@ -847,7 +850,7 @@ class Sequential(Model):
             steps_per_epoch=None,
             validation_steps=None,
             **kwargs):
-        """Trains the model for a fixed number of epochs.
+        """Trains the model for a fixed number of epochs (iterations on a dataset).
 
         # Arguments
             x: Numpy array of training data.
@@ -876,7 +879,7 @@ class Sequential(Model):
             callbacks: List of `keras.callbacks.Callback` instances.
                 List of callbacks to apply during training.
                 See [callbacks](/callbacks).
-            validation_split: Float between 0 and 1:
+            validation_split: Float between 0 and 1.
                 Fraction of the training data to be used as validation data.
                 The model will set apart this fraction of the training data,
                 will not train on it, and will evaluate
@@ -917,7 +920,7 @@ class Sequential(Model):
                 before declaring one epoch finished and starting the
                 next epoch. When training with input tensors such as
                 TensorFlow data tensors, the default `None` is equal to
-                the number of unique samples in your dataset divided by
+                the number of samples in your dataset divided by
                 the batch size, or 1 if that cannot be determined.
             validation_steps: Only relevant if `steps_per_epoch`
                 is specified. Total number of steps (batches of samples)
@@ -959,17 +962,28 @@ class Sequential(Model):
                               steps_per_epoch=steps_per_epoch,
                               validation_steps=validation_steps)
 
-    def evaluate(self, x, y, batch_size=32, verbose=1,
-                 sample_weight=None):
+    def evaluate(self, x=None, y=None,
+                 batch_size=None,
+                 verbose=1,
+                 sample_weight=None,
+                 steps=None):
         """Computes the loss on some input data, batch by batch.
 
         # Arguments
             x: input data, as a Numpy array or list of Numpy arrays
                 (if the model has multiple inputs).
+                `x` can be `None` (default) if feeding from
+                framework-native tensors (e.g. TensorFlow data tensors).
             y: labels, as a Numpy array.
-            batch_size: integer. Number of samples per gradient update.
+                `y` can be `None` (default) if feeding from
+                framework-native tensors (e.g. TensorFlow data tensors).
+            batch_size: Integer. If unspecified, it will default to 32.
             verbose: verbosity mode, 0 or 1.
             sample_weight: sample weights, as a Numpy array.
+            steps: Integer or `None`.
+                Total number of steps (batches of samples)
+                before declaring the evaluation round finished.
+                Ignored with the default value of `None`.
 
         # Returns
             Scalar test loss (if the model has no metrics)
@@ -986,24 +1000,29 @@ class Sequential(Model):
         return self.model.evaluate(x, y,
                                    batch_size=batch_size,
                                    verbose=verbose,
-                                   sample_weight=sample_weight)
+                                   sample_weight=sample_weight,
+                                   steps=steps)
 
-    def predict(self, x, batch_size=32, verbose=0):
+    def predict(self, x, batch_size=None, verbose=0, steps=None):
         """Generates output predictions for the input samples.
 
         The input samples are processed batch by batch.
 
         # Arguments
             x: the input data, as a Numpy array.
-            batch_size: integer.
+            batch_size: Integer. If unspecified, it will default to 32.
             verbose: verbosity mode, 0 or 1.
+            steps: Total number of steps (batches of samples)
+                before declaring the prediction round finished.
+                Ignored with the default value of `None`.
 
         # Returns
             A Numpy array of predictions.
         """
         if not self.built:
             self.build()
-        return self.model.predict(x, batch_size=batch_size, verbose=verbose)
+        return self.model.predict(x, batch_size=batch_size, verbose=verbose,
+                                  steps=steps)
 
     def predict_on_batch(self, x):
         """Returns predictions for a single batch of samples.
@@ -1072,7 +1091,7 @@ class Sequential(Model):
         return self.model.test_on_batch(x, y,
                                         sample_weight=sample_weight)
 
-    def predict_proba(self, x, batch_size=32, verbose=1):
+    def predict_proba(self, x, batch_size=None, verbose=0, steps=None):
         """Generates class probability predictions for the input samples.
 
         The input samples are processed batch by batch.
@@ -1080,13 +1099,17 @@ class Sequential(Model):
         # Arguments
             x: input data, as a Numpy array or list of Numpy arrays
                 (if the model has multiple inputs).
-            batch_size: integer.
+            batch_size: Integer. If unspecified, it will default to 32.
             verbose: verbosity mode, 0 or 1.
+            steps: Total number of steps (batches of samples)
+                before declaring the prediction round finished.
+                Ignored with the default value of `None`.
+
 
         # Returns
             A Numpy array of probability predictions.
         """
-        preds = self.predict(x, batch_size, verbose)
+        preds = self.predict(x, batch_size, verbose, steps=steps)
         if preds.min() < 0. or preds.max() > 1.:
             warnings.warn('Network returning invalid probability values. '
                           'The last layer might not normalize predictions '
@@ -1094,7 +1117,7 @@ class Sequential(Model):
                           '(like softmax or sigmoid would).')
         return preds
 
-    def predict_classes(self, x, batch_size=32, verbose=1):
+    def predict_classes(self, x, batch_size=None, verbose=0, steps=None):
         """Generate class predictions for the input samples.
 
         The input samples are processed batch by batch.
@@ -1102,21 +1125,26 @@ class Sequential(Model):
         # Arguments
             x: input data, as a Numpy array or list of Numpy arrays
                 (if the model has multiple inputs).
-            batch_size: integer.
+            batch_size: Integer. If unspecified, it will default to 32.
             verbose: verbosity mode, 0 or 1.
+            steps: Total number of steps (batches of samples)
+                before declaring the prediction round finished.
+                Ignored with the default value of `None`.
 
         # Returns
             A numpy array of class predictions.
         """
-        proba = self.predict(x, batch_size=batch_size, verbose=verbose)
+        proba = self.predict(x, batch_size=batch_size, verbose=verbose,
+                             steps=steps)
         if proba.shape[-1] > 1:
             return proba.argmax(axis=-1)
         else:
             return (proba > 0.5).astype('int32')
 
     @interfaces.legacy_generator_methods_support
-    def fit_generator(self, generator,
-                      steps_per_epoch,
+    def fit_generator(self,
+                      generator,
+                      steps_per_epoch=None,
                       epochs=1,
                       verbose=1,
                       callbacks=None,
@@ -1134,52 +1162,77 @@ class Sequential(Model):
         For instance, this allows you to do real-time data augmentation
         on images on CPU in parallel to training your model on GPU.
 
+        The use of `keras.utils.Sequence` guarantees the ordering
+        and guarantees the single use of every input per epoch when
+        using `use_multiprocessing=True`.
+
         # Arguments
-            generator: A generator.
+            generator: A generator or an instance of `Sequence`
+                (`keras.utils.Sequence`) object in order to avoid duplicate data
+                    when using multiprocessing.
                 The output of the generator must be either
-                - a tuple (inputs, targets)
-                - a tuple (inputs, targets, sample_weights).
-                All arrays should contain the same number of samples.
+                - a tuple `(inputs, targets)`
+                - a tuple `(inputs, targets, sample_weights)`.
+                This tuple (a single output of the generator) makes a single
+                batch. Therefore, all arrays in this tuple must have the same
+                length (equal to the size of this batch). Different batches may
+                have different sizes. For example, the last batch of the epoch
+                is commonly smaller than the others, if the size of the dataset
+                is not divisible by the batch size.
                 The generator is expected to loop over its data
                 indefinitely. An epoch finishes when `steps_per_epoch`
                 batches have been seen by the model.
             steps_per_epoch: Total number of steps (batches of samples)
                 to yield from `generator` before declaring one epoch
                 finished and starting the next epoch. It should typically
-                be equal to the number of unique samples of your dataset
+                be equal to the number of samples of your dataset
                 divided by the batch size.
+                Optional for `Sequence`: if unspecified, will use
+                the `len(generator)` as a number of steps.
             epochs: Integer, total number of iterations on the data.
                 Note that in conjunction with initial_epoch, the parameter
                 epochs is to be understood as "final epoch". The model is
                 not trained for n steps given by epochs, but until the
                 epoch epochs is reached.
-            verbose: Verbosity mode, 0, 1, or 2.
-            callbacks: List of callbacks to be called during training.
+            verbose: Integer. 0, 1, or 2. Verbosity mode.
+                0 = silent, 1 = progress bar, 2 = one line per epoch.
+            callbacks: List of `keras.callbacks.Callback` instances.
+                List of callbacks to apply during training.
+                See [callbacks](/callbacks).
             validation_data: This can be either
-                - A generator for the validation data
-                - A tuple (inputs, targets)
-                - A tuple (inputs, targets, sample_weights).
+                - a generator for the validation data
+                - a tuple `(inputs, targets)`
+                - a tuple `(inputs, targets, sample_weights)`.
             validation_steps: Only relevant if `validation_data`
-                is a generator.
-                Number of steps to yield from validation generator
+                is a generator. Total number of steps (batches of samples)
+                to yield from `validation_data` generator before stopping
                 at the end of every epoch. It should typically
-                be equal to the number of unique samples of your
+                be equal to the number of samples of your
                 validation dataset divided by the batch size.
-            class_weight: Dictionary mapping class indices to a weight
-                for the class.
-            max_queue_size: Maximum size for the generator queue
-            workers: Maximum number of processes to spin up
-            use_multiprocessing: if True, use process based threading.
-                Note that because
-                this implementation relies on multiprocessing,
-                you should not pass
-                non picklable arguments to the generator
-                as they can't be passed
-                easily to children processes.
-            shuffle: Whether to shuffle the order of the batches at
+                Optional for `Sequence`: if unspecified, will use
+                the `len(validation_data)` as a number of steps.
+            class_weight: Optional dictionary mapping class indices (integers)
+                to a weight (float) value, used for weighting the loss function
+                (during training only). This can be useful to tell the model to
+                "pay more attention" to samples from an under-represented class.
+            max_queue_size: Integer. Maximum size for the generator queue.
+                If unspecified, `max_queue_size` will default to 10.
+            workers: Integer. Maximum number of processes to spin up
+                when using process-based threading.
+                If unspecified, `workers` will default to 1. If 0, will
+                execute the generator on the main thread.
+            use_multiprocessing: Boolean.
+                If `True`, use process-based threading.
+                If unspecified, `use_multiprocessing` will default to `False`.
+                Note that because this implementation relies on multiprocessing,
+                you should not pass non-picklable arguments to the generator
+                as they can't be passed easily to children processes.
+            shuffle: Boolean (whether to shuffle the order of the batches at
                 the beginning of each epoch. Only used with instances
-                of `Sequence` (keras.utils.Sequence).
-            initial_epoch: Epoch at which to start training
+                of `Sequence` (`keras.utils.Sequence`).
+                Has no effect when `steps_per_epoch` is not `None`.
+            initial_epoch: Integer.
+                Epoch at which to start training
                 (useful for resuming a previous training run).
 
         # Returns
@@ -1187,19 +1240,19 @@ class Sequential(Model):
 
         # Raises
             RuntimeError: if the model was never compiled.
+            ValueError: In case the generator yields data in an invalid format.
 
         # Example
 
         ```python
             def generate_arrays_from_file(path):
-                while 1:
-                    f = open(path)
-                    for line in f:
-                        # create Numpy arrays of input data
-                        # and labels, from each line in the file
-                        x, y = process_line(line)
-                        yield (x, y)
-                    f.close()
+                while True:
+                    with open(path) as f:
+                        for line in f:
+                            # create Numpy arrays of input data
+                            # and labels, from each line in the file
+                            x, y = process_line(line)
+                            yield (x, y)
 
             model.fit_generator(generate_arrays_from_file('/my_file.txt'),
                                 steps_per_epoch=1000, epochs=10)
@@ -1223,7 +1276,7 @@ class Sequential(Model):
                                         initial_epoch=initial_epoch)
 
     @interfaces.legacy_generator_methods_support
-    def evaluate_generator(self, generator, steps,
+    def evaluate_generator(self, generator, steps=None,
                            max_queue_size=10, workers=1,
                            use_multiprocessing=False):
         """Evaluates the model on a data generator.
@@ -1236,6 +1289,8 @@ class Sequential(Model):
                 or (inputs, targets, sample_weights)
             steps: Total number of steps (batches of samples)
                 to yield from `generator` before stopping.
+                Optional for `Sequence`: if unspecified, will use
+                the `len(generator)` as a number of steps.
             max_queue_size: maximum size for the generator queue
             workers: maximum number of processes to spin up
             use_multiprocessing: if True, use process based threading.
@@ -1263,7 +1318,7 @@ class Sequential(Model):
                                              use_multiprocessing=use_multiprocessing)
 
     @interfaces.legacy_generator_methods_support
-    def predict_generator(self, generator, steps,
+    def predict_generator(self, generator, steps=None,
                           max_queue_size=10, workers=1,
                           use_multiprocessing=False, verbose=0):
         """Generates predictions for the input samples from a data generator.
@@ -1275,6 +1330,8 @@ class Sequential(Model):
             generator: generator yielding batches of input samples.
             steps: Total number of steps (batches of samples)
                 to yield from `generator` before stopping.
+                Optional for `Sequence`: if unspecified, will use
+                the `len(generator)` as a number of steps.
             max_queue_size: maximum size for the generator queue
             workers: maximum number of processes to spin up
             use_multiprocessing: if True, use process based threading.
@@ -1466,10 +1523,10 @@ def _clone_functional_model(model, input_tensors=None):
         tensor_map[x] = (y, None)  # tensor, mask
 
     # Iterated over every node in the reference model, in depth order.
-    depth_keys = list(model.nodes_by_depth.keys())
+    depth_keys = list(model._nodes_by_depth.keys())
     depth_keys.sort(reverse=True)
     for depth in depth_keys:
-        nodes = model.nodes_by_depth[depth]
+        nodes = model._nodes_by_depth[depth]
         for node in nodes:
             # Recover the corresponding layer.
             layer = node.outbound_layer
